      SUBROUTINE crs_o3(crs, error, first, lambda, n_lambda, ntemp, 
     $     path, temper, type)  
*
*
*BOD
*DESCRIPTION
* Calculates ozone cross sections as a function of
* wavelength and temperature.
*USAGE
**FORTRAN
*     CALL crs_o3( crs, error, first, lambda, n_lambda, ntemp, path,
*    $             temper, type)  
**PERL
*INPUT
**first
*     logical
*      If true, temperature coefficients for ozone will be 
*      calculated at wavelengths lambda. If first is true, it will
*      be returned as false, just in case you should forget.
*      If false, the temperature coefficients from earlier call will be
*      used. 
*      first should be true the first time crs_o3 is called. It should 
*      also be true if lambda is changed. It does not need to be true
*      if temper is changed.
**lambda
*     real array(*)
*     Wavelengths (nm) at which the ozone cross section is to be calculated.
**n_lambda
*     integer
*     Number of wavelengths lambda.
**ntemp
*     integer
*     Number of different temperatures (levels in the atmosphere) 
**path
*     character string
*     The location of the Radtran data files, e.g. 
*     /usr/local/lib/Radtran/data
**temper
*     real array(*)
*     The temperatures (levels in the atmosphere) at which
*     the ozone cross section is to be calculated.
**type
*     integer
*     The ozone cross section data to be used:
***1     Molina and Molina (1986)
***2     Daumont et al. (1992), Malicet et al. (1995)
***3     Paur and Bass (1985)
*OUTPUT
**crs
*     real array(temper,lambda)
*     The ozone cross section as a function of temper and lambda.
**error
*     integer
*     Error flag. Error checking performed only if first is true.
***0     Everything ok.
***1     Error occurred during reading of data file.
***2     n_lambda == 0 and ntemp != 0
***3     n_lambda != 0 and ntemp == 0
*ERROR_MESSAGES
*ROUTINES_CALLED
*FILES
*     crs_o3_all.dat           data file
*     crs_o3_dau.dat           data file
*     crs_o3_pab.dat           data file
*     consts.inc               Include file
*     param.inc                Include file
*REFERENCES
*     (1) Molina and Molina, 'Absolute Absorption Cross Sections of
*         Ozone in the 185- to 350-nm wavelength Range', JGR, vol 91,
*         pp. 14,501-14,508, 1986.
*
*     (2) Daumont D, Brion J, Charbonnier J and Malicet J, 
*         "Absorption Cross-Section at Room Temperature",
*         J.Atmos.Chem. no. 15, vol. 145 (1992)
*
*         Malicet J, Daumont D, Charbonnier J, Parisse C, Chakir A, 
*         and Brion J, Ozone UV spectroscopy II: Absorption Cross-Sections
*         and Temperature Dependence, J. Atmos. Chem., 21 (1995), 263-273
*
*     (3) Paur and Bass, (1985) 'The ultraviolet cross sections of ozone, II, 
*         results and temperature dependence', in: Atmospheric Ozone, proceedings 
*         of the Quadrennial Ozone Symposium, Halkidiki, Greece. 
*         C. Zerefos and A. Ghazi (eds), pp.611-616, D. Reidel, Hingham, MA 
*
*EOD
*
*
*     References:
*     
*     WMO (1986), 'Atmospheric Ozone 1985', WMO Report No. 16.
*
*     
      INCLUDE 'param.inc'
*
      CHARACTER*(*) path
      INTEGER  error,  n_lambda, ntemp, type
      LOGICAL first
*
      REAL crs(0:maxcly, maxpro, *), lambda(*), temper(0:*)
*
* Internal variables
*     
      CHARACTER*1  char
      CHARACTER*18 filenm    ! No. of characters must equal
                             ! length of filename.
      CHARACTER*100 fullname, line
      INTEGER n_temp, n_temp_raw(maxwvn), n_crs
      REAL crs_temp(NCRSTEMP), crs_raw(NCRSTEMP,maxwvnraw), fact,
     $     lambda_raw(maxwvnraw), crs_new(NCRSTEMP),
     $     temp_new(NCRSTEMP,maxwvn), coeff(NCRSTEMP, maxwvn),
     $     coeff_crs(NCRSTEMP, maxwvnraw)
      real a, b, gradient

*
      INTEGER closest_above, closest_below
*
      real lambda_vac(maxwvnraw), lambda_air(maxwvnraw)
*
      SAVE  n_temp, n_crs, crs_temp, crs_raw, lambda_raw, coeff,
     $     n_temp_raw, temp_new
*
      INCLUDE 'consts.inc'
*
      if (type .eq. 3) then 
*     This option pertains to a case where the o3 cross-sections are 
*     given from a fitted formula to data taken by Bass and Paur (1985)
*     
*     Calculate ozone cross sections 
*     

         CALL zeroit( crs, (maxcly+1)*maxwvn*maxpro)
*
         IF ( ( n_lambda .LE. 0 ) .AND. ( ntemp .NE. 0 ) ) THEN
            error = 2
            RETURN
         ELSE IF ( ( n_lambda .LE. 0 ) .AND. ( ntemp .NE. 0 ) ) THEN
            error = 3
            RETURN
         ENDIF
*
         filenm   = 'crs/crs_o3_pab.dat'
         n_coeff = 3
         n_temp   = 3
         n_crs    = 1
         fact     = 1.E-20
*     
         fullname = path//filenm
         OPEN(UNIT=iunit,FILE=fullname,STATUS='OLD',
     $        FORM='FORMATTED',ERR=89, IOSTAT=ios)
         GOTO 881
 89      CONTINUE
         WRITE(*,'(A,A)') 'Error during read of file: ', fullname
         WRITE(*,'(A, i4)') 'IOSTAT: ', ios
         error = 1
         RETURN
 881     CONTINUE
*     
         DO WHILE ( .TRUE. ) 
            READ(iunit,'(A)',END=88 ) line
*     
*     Read all the no-comment lines
*     
            IF ( line(:1) .NE. comcha ) THEN
               READ(line,*) lambda_vac(n_crs),   
     $              (coeff_crs(it, n_crs),it=1, n_temp), 
     $              lambda_air(n_crs)
*               write(*,*) n_crs, lambda_vac(n_crs),   
*     $              (coeff_crs(it, n_crs),it=1,n_temp),
*     $              lambda_air(n_crs)
               n_crs = n_crs + 1
            ENDIF
         ENDDO
 88      CONTINUE
         n_crs = n_crs - 1
         CLOSE(iunit)

         do iv=1, n_lambda
            ia=closest_above(lambda(iv), lambda_air, n_crs)
            ib=closest_below(lambda(iv), lambda_air, n_crs)
            if ((ia .gt. n_crs) .and. (ib .eq. 0)) then
               write(*,*) 'crs_o3, cannot handle this case yet'
               write(*,*) 'lambda(iv), n_crs, ia, ib'
               write(*,*) lambda(iv), n_crs, ia, ib
               stop
            endif
            DO it = 0, ntemp
               a=(coeff_crs(1,ia) + 
     $              coeff_crs(2,ia)*(temper(it)-273.15) + 
     $              coeff_crs(3,ia) * (temper(it)-273.15)**2.)
               if (ia .eq. ib) then
                  crs(it, io3, iv) = a*fact
               else 
                  b=(coeff_crs(1,ib) + 
     $                 coeff_crs(2,ib)*(temper(it)-273.15) + 
     $                 coeff_crs(3,ib) * (temper(it)-273.15)**2.)
                  gradient=(a-b)/(lambda_air(ia)-lambda_air(ib))
                  crs(it, io3, iv) =  
     $                 (gradient*(lambda(iv)-lambda_air(ib))+b) * fact
               endif
            ENDDO
         ENDDO
         RETURN
      ENDIF


*
      error = 0

*
      IF ( first ) THEN
*
         CALL zeroit( crs, (maxcly+1)*maxwvn*maxpro)
*
         IF ( ( n_lambda .LE. 0 ) .AND. ( ntemp .NE. 0 ) ) THEN
            error = 2
            RETURN
         ELSE IF ( ( n_lambda .LE. 0 ) .AND. ( ntemp .NE. 0 ) ) THEN
            error = 3
            RETURN
         ENDIF
*
* Read in all data on first entry only
*
         IF ( type .EQ. 1 ) THEN
*            filenm   = 'crs/crs_o3_Molina.dat'
            filenm   = 'crs/crs_o3_all.dat'
            n_temp   = 3
            n_crs    = 1
            fact     = 1.E-20
         ENDIF
*
         IF ( type .EQ. 2 ) THEN
            filenm   = 'crs/crs_o3_dau.dat'
            n_temp   = 4
            n_crs    = 1
            fact     = 1.E-20
         ENDIF
*
         fullname = path//filenm
         OPEN(UNIT=iunit,FILE=fullname,STATUS='OLD',
     $        FORM='FORMATTED',ERR=99, IOSTAT=ios)
         GOTO 991
 99      CONTINUE
         WRITE(*,'(A,A)') 'Error during read of file: ', fullname
         WRITE(*,'(A, i4)') 'IOSTAT: ', ios
         error = 1
         RETURN
 991     CONTINUE
*     
         DO WHILE ( .TRUE. ) 
            READ(iunit,'(A)', END=98 ) line
*
* Read all the no-comment lines
*
            IF ( line(:1) .NE. comcha ) THEN
               IF ((type .EQ. 1) .or. (type .eq. 2)) THEN
                  IF ( line(:1) .EQ. tempcha ) THEN
                     READ(line,*) char, (crs_temp(it),it=1,n_temp)
                  ELSE
                     READ(line,*) lambda_raw(n_crs),   
     $                    (crs_raw(it, n_crs),it=1,n_temp)
*                     write(*,*) lambda_raw(n_crs),   
*     $                    (crs_raw(it, n_crs),it=1,n_temp)
                     DO it=1, n_temp
                        crs_raw(it, n_crs) = crs_raw(it, n_crs) * fact
                     ENDDO
                     n_crs = n_crs + 1
                  ENDIF
               ENDIF
            ENDIF
         ENDDO
 98      CONTINUE
         n_crs = n_crs - 1
         CLOSE(iunit)
*
* Return raw cross section if n_lambda and ntemp equal zero
*
         IF ( ( n_lambda .EQ. 0 ) .AND. ( ntemp .EQ. 0 ) ) THEN
            DO it = 0, n_temp
               temper(it) = crs_temp(it)
               DO iv = 1, n_crs
                  lambda(iv)  = lambda_raw(iv)
                  crs(it, io3, iv) = crs_raw(it,iv)
               ENDDO
            ENDDO
            n_lambda = n_crs
            ntemp    = n_temp
            RETURN
         ENDIF
*
* Calculate the temperature coefficients for all lambda.
*
         DO iv = 1, n_lambda
*
*  Find the closest wavelengths in original data.
*
*
            ia = closest_above(lambda(iv), lambda_raw, n_crs)
            ib = closest_below(lambda(iv), lambda_raw, n_crs)
*            write(*,*) "doing closest", iv, lambda(iv), ia,ib,
*     $           crs_raw(1,ia), crs_raw(1,ib)
            IF ((ib .EQ. 0) .OR. (ia .GT. n_crs)) THEN 
               WRITE(*,*) "crs_o3, cannot handle this case yet"
            ELSE
*     
*  Figure out at how many temperatures the cross section 
*  is reported.
*
*     Molina and Molina report values mostly at three but 
*        sometimes only at two temperatures.
               IF ((type .EQ. 1) .AND. 
     $              ((crs_raw(2,ia) .EQ. 0.0 ) .OR.
     $              (crs_raw(2,ib) .EQ. 0.0 ))) THEN
                  n_temp_raw(iv) = 2
                  if (ia .eq. ib) then
                     crs_new(1)=crs_raw(1,ia)
                     crs_new(2)=crs_raw(3,ia)
                  else 
                     a = crs_raw(1,ia)
                     b = crs_raw(1,ib)
                     gradient = (a-b)/(lambda_raw(ia)-lambda_raw(ib))
                     crs_new(1) = gradient * (lambda(iv)-lambda_raw(ib))
     $                    + b
                     a = crs_raw(3,ia)
                     b = crs_raw(3,ib)
                     gradient = (a-b)/(lambda_raw(ia)-lambda_raw(ib))
                     crs_new(2) = gradient * (lambda(iv)-lambda_raw(ib))
     $                    + b
                  endif
                  temp_new(1,iv)= crs_temp(1)
                  temp_new(2,iv)= crs_temp(3)
               ELSE IF ( type .EQ. 1 ) THEN
                  n_temp_raw(iv) = 3
                  a = crs_raw(1,ia)
                  if (ia .eq. ib) then
                     crs_new(1)=crs_raw(1,ia)
                     crs_new(2)=crs_raw(2,ia)
                     crs_new(3)=crs_raw(3,ia)
                  else 
                     b = crs_raw(1,ib)
                     gradient = (a-b)/(lambda_raw(ia)-lambda_raw(ib))
                     crs_new(1) = gradient * (lambda(iv)-lambda_raw(ib))
     $                    + b
                     a = crs_raw(2,ia)
                     b = crs_raw(2,ib)
                     gradient = (a-b)/(lambda_raw(ia)-lambda_raw(ib))
                     crs_new(2) = gradient * (lambda(iv)-lambda_raw(ib))
     $                    + b
                     a = crs_raw(3,ia)
                     b = crs_raw(3,ib)
                     gradient = (a-b)/(lambda_raw(ia)-lambda_raw(ib))
                     crs_new(3) = gradient * (lambda(iv)-lambda_raw(ib))
     $                    + b
                  endif
                  temp_new(1,iv)= crs_temp(1)
                  temp_new(2,iv)= crs_temp(2)
                  temp_new(3,iv)= crs_temp(3)
cakyp
*                  WRITE(*,*)"CRS",iv,lambda(iv), gradient,
*     $                 lambda_raw(ia),lambda_raw(ib), a, b 
*                  WRITE(*,*)"CRS",iv,lambda(iv),crs_new(1),
*     $                 crs_new(2),crs_new(3),
*     $                 temp_new(1,iv),temp_new(2,iv),temp_new(3,iv)
cakyp
               ENDIF
*     Malicet and Daumont et al. report values at both 
*     three and four temperatures.
               IF ((type .EQ. 2) .AND. 
     $              ((crs_raw(3,ia) .EQ. 0.0 ) .OR.
     $              (crs_raw(3,ib) .EQ. 0.0 ))) THEN
                  n_temp_raw(iv) = 3
                  if (ia .eq. ib) then
                     crs_new(1)=crs_raw(1,ia)
                     crs_new(2)=crs_raw(2,ia)
                     crs_new(3)=crs_raw(3,ia)
                  else 
                     a = crs_raw(1,ia)
                     b = crs_raw(1,ib)
                     gradient = (a-b)/(lambda_raw(ia)-lambda_raw(ib))
                     crs_new(1) = gradient * (lambda(iv)-lambda_raw(ib))
     $                    + b
                     a = crs_raw(2,ia)
                     b = crs_raw(2,ib)
                     gradient = (a-b)/(lambda_raw(ia)-lambda_raw(ib))
                     crs_new(2) = gradient * (lambda(iv)-lambda_raw(ib))
     $                    + b
                     a = crs_raw(3,ia)
                     b = crs_raw(3,ib)
                     gradient = (a-b)/(lambda_raw(ia)-lambda_raw(ib))
                     crs_new(3) = gradient * (lambda(iv)-lambda_raw(ib))
     $                    + b
                  endif
                  temp_new(1,iv)= crs_temp(1)
                  temp_new(2,iv)= crs_temp(2)
                  temp_new(3,iv)= crs_temp(4)
               ELSE IF ( type .EQ. 2 ) THEN
                  n_temp_raw(iv) = 4
                  if (ia .eq. ib) then
                     crs_new(1)=crs_raw(1,ia)
                     crs_new(2)=crs_raw(2,ia)
                     crs_new(3)=crs_raw(3,ia)
                     crs_new(4)=crs_raw(4,ia)
                  else 
                     a = crs_raw(1,ia)
                     b = crs_raw(1,ib)
                     gradient = (a-b)/(lambda_raw(ia)-lambda_raw(ib))
                     crs_new(1) = gradient * (lambda(iv)-lambda_raw(ib))
     $                    + b
                     a = crs_raw(2,ia)
                     b = crs_raw(2,ib)
                     gradient = (a-b)/(lambda_raw(ia)-lambda_raw(ib))
                     crs_new(2) = gradient * (lambda(iv)-lambda_raw(ib))
     $                    + b
                     a = crs_raw(3,ia)
                     b = crs_raw(3,ib)
                     gradient = (a-b)/(lambda_raw(ia)-lambda_raw(ib))
                     crs_new(3) = gradient * (lambda(iv)-lambda_raw(ib))
     $                    + b
                     a = crs_raw(4,ia)
                     b = crs_raw(4,ib)
                     gradient = (a-b)/(lambda_raw(ia)-lambda_raw(ib))
                     crs_new(4) = gradient * (lambda(iv)-lambda_raw(ib))
     $                    + b
                  endif
                  temp_new(1,iv)= crs_temp(1)
                  temp_new(2,iv)= crs_temp(2)
                  temp_new(3,iv)= crs_temp(3)
                  temp_new(4,iv)= crs_temp(4)
cakyp
*     WRITE(*,*)"CRS", iv, lambda(iv), crs_new(1),
*     $                 crs_new(2), crs_new(3),
*     $                 temp_new(1,iv), temp_new(2,iv), temp_new(3,iv)
cakyp

               ENDIF
            ENDIF
cakyp            write(*,*) "doing newton", iv, lambda(iv), maxwvn, maxwvnraw
*
*  Use these cross sections to calculate temperature coefficients.
*
            CALL newton_co(n_temp_raw(iv), temp_new(1,iv), crs_new, 
     $           coeff(1,iv))
cakyp
*                  WRITE(*,*)"coeff", iv, lambda(iv), coeff(1,iv)
cakyp
         ENDDO
*
         first = .FALSE.
      ENDIF
*
********** Done with first ************************************
*
* Below follows what is done on .NOT. first calls
*
* Calculate ozone cross sections 
*
      DO iv = 1, n_lambda
         DO it = 0, ntemp
            crs(it, io3, iv) = eval(n_temp_raw(iv), temp_new(1,iv),
     $           coeff(1,iv), temper(it))
         ENDDO
*
*         it = ntemp
*            write(*,*) "loop crs_o3", iv, it, n_lambda, ntemp,
*     $      lambda(iv), crs(it, io3, iv), 
*     $        coeff(1,iv), coeff(2,iv), coeff(3,iv),
*     $        n_temp_raw(iv), temp_new(1,iv), temp_new(2,iv), 
*     $        temp_new(3,iv),
*     $        temper(it)
*
      ENDDO
*
      RETURN
      END
      

