/************************************************************************/
/* stamnes.c                                                            */
/*                                                                      */
/* Library for Stamnes table related calculations.                      */
/*                                                                      */
/* Author: Bernhard Mayer,                                              */
/*        Fraunhofer Institute for Atmospheric Environmental Research,  */
/*        82467 Garmisch-Partenkirchen,                                 */
/*        Germany                                                       */
/*                                                                      */
/*----------------------------------------------------------------------*/
/* Copyright (C) 1997 Bernhard Mayer                                    */
/*                                                                      */
/* This program is free software; you can redistribute it and/or modify */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation; either version 1, or (at your option)  */
/* any later version.                                                   */
/*                                                                      */
/* This program is distributed in the hope that it will be useful,      */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY of FITNESS FOR A PARTICULAR PURPOSE. See the         */
/* GNU General Public License for more details.                         */
/*                                                                      */
/* To obtain a copy of the GNU General Public License write to the      */
/* Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,   */
/* USA.                                                                 */
/*----------------------------------------------------------------------*/
/************************************************************************/


#include <stamnes.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <process.h>
#include <float.h>
#include <math.h>
#include <string.h>
#include <numeric.h>
#include <ascii.h>
#include <dbase.h>
#include <sbase.h>
#include <sun.h>
#include <getopt.h>
#include <model.h>


#define NUMBER_OF_ZENITH     50
#define NUMBER_OF_RATIO     173

#define OZONE_LOWER_LIMIT     0.0
#define OZONE_UPPER_LIMIT   800.0
#define OZONE_STEP           50.0


/* Error codes */
#define NOT_A_VALID_STAMNES_TABLE -30
#define RATIO_OUT_OF_RANGE        -31
#define ERROR_SPLINING_ZENITH     -32


/* array of fix values to be listed in table */
/* zenith MUST not be 60.0 (DISORT problem!) */

double stamnes_zenith[NUMBER_OF_ZENITH] = {  0.01, 5,10,15,
				 	    20,22,24,26,28,
					    30,32,34,36,38,
					    40,42,44,46,48,
					    50,51,52,53,54,55,56,57,58,59,
					    60.01,61,62,63,64,65,66,67,68,69,
					    70,71,72,73,74,75,76,77,78,79,80}; 


double stamnes_ratio [NUMBER_OF_RATIO] = {5,5.1,5.2,5.3,5.4,5.5,5.6,5.7,5.8,5.9,
					  6,6.1,6.2,6.3,6.4,6.5,6.6,6.7,6.8,6.9,
					  7,7.1,7.2,7.3,7.4,7.5,7.6,7.7,7.8,7.9,
					  8,8.1,8.2,8.3,8.4,8.5,8.6,8.7,8.8,8.9,
					  9,9.1,9.2,9.3,9.4,9.5,9.6,9.7,9.8,9.9,
					  10,10.5,11,11.5,12,12.5,13,13.5,14,14.5,
					  15,15.5,16,16.5,17,17.5,18,18.5,19,19.5,
					  20,20.5,21,21.5,22,22.5,23,23.5,24,24.5,
					  25,25.5,26,26.5,27,27.5,28,28.5,29,29.5,
					  30,31,32,33,34,35,36,37,38,39,40,41,42,
					  43,44,45,46,47,48,49,50,55,60,65,70,75,80,
					  85,90,95,100,105,110,115,120,125,130,
					  135,140,145,150,155,160,165,170,175,
					  180,185,190,195,200,210,220,230,240,
					  250,260,270,280,290,300,310,320,330,
					  340,350,360,370,380,390,400,410,420,
					  430,440,450,460,470,480,490,500,550,600};




/************************************************************************/
/* Create a Stamnes table for the specified atmosphere, model type, ... */
/************************************************************************/

int create_stamnes_table (char *stafilename,
			  atmospheric_data *atm,
			  MODEL model,
			  RESOLUTION resolution,
			  STYLE style,
			  char *slitfunction,
			  double lambda_lower,
			  double lambda_upper,
			  char conv,
			  char diffuse)
{
  int status=0, i=0, j=0, k=0, l=0;

  double *ozone=0, *ratio=0;
  double value_lower=0, value_upper=0;
  double zenith=0;

  double *a0=NULL, *a1=NULL, *a2=NULL, *a3=NULL;
  double ozonevalue=0;

  double *global=NULL, *direct=NULL;
  double *conv_lambda=NULL, *conv_global=NULL;
  int n_ozone=0;

  double *lambda=NULL;

  double *x_conv=NULL, *y_conv=NULL;
  int n_conv=0;

  int n_global=0;

  /* array of tabulated values */
  int table[NUMBER_OF_RATIO][NUMBER_OF_ZENITH];

  FILE *stafile=NULL;



  n_ozone = (OZONE_UPPER_LIMIT - OZONE_LOWER_LIMIT) / OZONE_STEP;
  ozone = (double *) calloc (n_ozone, sizeof(double));
  ratio = (double *) calloc (n_ozone, sizeof(double));

  /* set ozone */
  for (j=0; j<n_ozone; j++)  
    ozone[j] = OZONE_LOWER_LIMIT + (double) j * OZONE_STEP;
    

  /* calculate convolution function */
  if (conv)  {
    n_conv=21;
    x_conv = calloc(n_conv, sizeof(double));
    y_conv = calloc(n_conv, sizeof(double));

    x_conv[0]  = -5.0;
    x_conv[1]  = -4.5;
    x_conv[2]  = -4.0;
    x_conv[3]  = -3.5;
    x_conv[4]  = -3.0;
    x_conv[5]  = -2.5;
    x_conv[6]  = -2.0;
    x_conv[7]  = -1.5;
    x_conv[8]  = -1.0;
    x_conv[9]  = -0.5;
    x_conv[10] =  0.0;
    x_conv[11] =  0.5;
    x_conv[12] =  1.0;
    x_conv[13] =  1.5;
    x_conv[14] =  2.0;
    x_conv[15] =  2.5;
    x_conv[16] =  3.0;
    x_conv[17] =  3.5;
    x_conv[18] =  4.0;
    x_conv[19] =  4.5;
    x_conv[20] =  5.0;

    y_conv[0]  =  0.0;
    y_conv[1]  =  0.1;
    y_conv[2]  =  0.2;
    y_conv[3]  =  0.3;
    y_conv[4]  =  0.4;
    y_conv[5]  =  0.5;
    y_conv[6]  =  0.6;
    y_conv[7]  =  0.7;
    y_conv[8]  =  0.8;
    y_conv[9]  =  0.9;
    y_conv[10] =  1.0;
    y_conv[11] =  0.9;
    y_conv[12] =  0.8;
    y_conv[13] =  0.7;
    y_conv[14] =  0.6;
    y_conv[15] =  0.5;
    y_conv[16] =  0.4;
    y_conv[17] =  0.3;
    y_conv[18] =  0.2;
    y_conv[19] =  0.1;
    y_conv[20] =  0.0;

    lambda=(double *) calloc (n_conv, sizeof(double));
  }
  else  {
    lambda=(double *) calloc (1, sizeof(double));
  }

    

  /* do for each zenith angle */
  for (i=0; i<NUMBER_OF_ZENITH; i++)  {
    zenith = stamnes_zenith[i];

    /* do for each ozone concentration */
    for (j=0; j<n_ozone; j++)  {

      /* set ozone column */
      atm->ozone = ozone[j];


      /* calculate model ratio upper / lower */


      if (!conv)  {   /* classic */

	/* call model for upper wavelength */
	lambda[0] = lambda_upper;
	

	/* call model */
	switch(style) {
	case OLD:
	  status = model_calculate (model, resolution,
				    atm,
				    lambda, 1, 
				    zenith, 1,
				    &global, &direct);
	  break;

	case BETTER:
	  status = better_model_calculate (model, resolution, slitfunction,
					   atm,
					   lambda, 1, 
					   zenith, 1,
					   &global, &direct);
	  break;

	default:
	  break;
	}


	/* if error occurred */
	if (status!=0)  {
	  fprintf (stderr, "model_calculate returned status %d\n", status);
	  return status;
	}
	
	/* check for correct wavelength */
	if (!double_equal (lambda_upper, lambda[0]))  {
	  fprintf (stderr, " ... fatal error!\n");
	  fprintf (stderr, "     model_calculate()  %g nm\n", lambda[0]);
	  fprintf (stderr, "     stamntab           %g nm\n", lambda_upper);
	  return (-1);
	}
	
	if (diffuse)
	  value_upper = global[0] - direct[0];
	else
	  value_upper = global[0];
	
	free(global);
	free(direct);
	
	
	/* call model for lower wavelength */
	lambda[0] = lambda_lower;
	

	/* call model */
	switch(style) {
	case OLD:
	  status = model_calculate (model, resolution,
				    atm,
				    lambda, 1, 
				    zenith, 1,
				    &global, &direct);
	  break;

	case BETTER:
	  status = better_model_calculate (model, resolution, slitfunction,
					   atm,
					   lambda, 1, 
					   zenith, 1,
					   &global, &direct);
	  break;

	default:
	  break;
	}


	if (status!=0)  { 
	  fprintf (stderr, "model_calculate returned status %d\n", status);
	  return status;
	}
	
	/* check for correct wavelength */
	if (!double_equal (lambda_lower, lambda[0]))  {
	  fprintf (stderr, "fatal error!\n");
	  return (-1);
	}
	
	if (diffuse)
	  value_lower = global[0] - direct[0];
	else
	  value_lower = global[0];


	free(global);
	free(direct);

      }

      else  {    /* convolute with 5nm FWHM triangle */
	
	/* call model for upper wavelength */

	for (k=0; k<n_conv; k++)
	  lambda[k] = lambda_upper + x_conv[k];
	

	/* call model */
	switch(style) {
	case OLD:
	  status = model_calculate (model, resolution,
				    atm,
				    lambda, n_conv, 
				    zenith, 1,
				    &global, &direct);
	  break;

	case BETTER:
	  status = better_model_calculate (model, resolution, slitfunction,
					   atm,
					   lambda, n_conv, 
					   zenith, 1,
					   &global, &direct);
	  break;

	default:
	  break;
	}


	
	/* if error occurred */
	if (status!=0)  {
	  fprintf (stderr, "model_calculate returned status %d\n", status);
	  return status;
	}
	
	if (diffuse)       /* replace global by diffuse */
	  for (l=0; l<n_conv; l++)
	    global[l] -= direct[l];


	/* convolute with triangle */
	status = convolute (lambda, global, n_conv,
			    x_conv, y_conv, n_conv,
			    &conv_lambda, &conv_global, &n_global);
	
	if (status!=0)  {
	  fprintf (stderr, " ... FATAL error convoluting global\n");
	  return status;
	}
	  
	

	/* check for correct wavelength */
	if (!double_equal (lambda_upper, conv_lambda[(n_conv-1)/2]))  {
	  fprintf (stderr, " ... fatal error!\n");
	  fprintf (stderr, "     model_calculate()  %g nm\n", conv_lambda[(n_conv-1)/2]);
	  fprintf (stderr, "     stamntab           %g nm\n", lambda_upper);
	  return (-1);
	}
	
	
	value_upper = conv_global[(n_conv-1)/2];

	free(conv_lambda);
	free(conv_global);
	free(global);
	free(direct);
	
	

	/* call model for lower wavelength */

	for (k=0; k<n_conv; k++)
	  lambda[k] = lambda_lower + x_conv[k];
	

	/* call model */
	switch(style) {
	case OLD:
	  status = model_calculate (model, resolution,
				    atm,
				    lambda, n_conv, 
				    zenith, 1,
				    &global, &direct);
	  break;

	case BETTER:
	  status = better_model_calculate (model, resolution, slitfunction,
					   atm,
					   lambda, n_conv, 
					   zenith, 1,
					   &global, &direct);
	  break;

	default:
	  break;
	}

	
	/* if error occurred */
	if (status!=0)  {
	  fprintf (stderr, "model_calculate returned status %d\n", status);
	  return status;
	}
	
	if (diffuse)       /* replace global by diffuse */
	  for (l=0; l<n_conv; l++)
	    global[l] -= direct[l];


	/* convolute with triangle */
	status = convolute (lambda, global, n_conv,
			    x_conv, y_conv, n_conv,
			    &conv_lambda, &conv_global, &n_global);

	if (status!=0)  {
	  fprintf (stderr, " ... FATAL error convoluting\n");
	  return status;
	}
	  
	
	/* check for correct wavelength */
	if (!double_equal (lambda_lower, conv_lambda[(n_conv-1)/2]))  {
	  fprintf (stderr, " ... fatal error!\n");
	  fprintf (stderr, "     model_calculate()  %g nm\n", conv_lambda[(n_conv-1)/2]);
	  fprintf (stderr, "     stamntab           %g nm\n", lambda_lower);
	  return (-1);
	}
	
	value_lower = conv_global[(n_conv-1)/2];
	
	free(conv_lambda);
	free(conv_global);
	free(global);
	free(direct);
	
      }
	  
      /* calculate ratio */
      ratio[j] = value_upper / value_lower;
	
      fprintf (stderr, " ... calculated ratio for zenith = %g, ozone = %g: %g\n", 
	       zenith, atm->ozone, ratio[j]);
	      
    }

    status = spline_coeffc (ratio, ozone, n_ozone, &a0, &a1, &a2, &a3);
    if (status != 0)  {
      fprintf (stderr, "error calculating spline coefficients\n");
      return (-1);
    }

    /* spline (ratio[], ozone[]) to stamnes_ratio[] and save to table[][i] */
    for (j=0; j<NUMBER_OF_RATIO; j++)  {

      status = calc_splined_value (stamnes_ratio[j], &ozonevalue, 
				   ratio, n_ozone, a0, a1, a2, a3);
      if (status == 0) 
	table[j][i] = (int) ozonevalue;
      else
	table[j][i] = -1;
    }

    free(a0);
    free(a1);
    free(a2);
    free(a3);

  } 



  /* print table to file, including header */
  
  /* open output file */
  if ((stafile = fopen(stafilename, "w")) == NULL)  {
    fprintf (stderr, "error opening output file %s\n", stafilename);
    return (-1);
  }


  /* print header line */
  fprintf (stafile, "%6.1f ", -1.0);
  for (j=0; j<NUMBER_OF_ZENITH; j++) 
    fprintf (stafile, "%5.0f", stamnes_zenith[j]);

  fprintf (stafile, "\n");
  
  for (i=0; i<NUMBER_OF_RATIO; i++)  {
    fprintf (stafile, "%6.1f ", stamnes_ratio[i]);
    for (j=0; j<NUMBER_OF_ZENITH; j++)  
      fprintf (stafile, "%5d", table[i][j]);
    
    fprintf (stafile, "\n");
  }
  
  fclose (stafile);
  return 0;   /* if o.k. */
}







/**************************************************************/
/* Read Stamnes table from file filename.                     */
/* zenith [0..n_zenith-1] is an array of solar zenith angles, */ 
/* ratio [0..n_ratio-1] an array of ratios found in the       */
/* file. The two-dimensional integer array ozone holds the    */
/* corresponding ozone columns.                               */
/* Memory for arrays is allocated automatically.              */
/**************************************************************/

int read_stamnes_table (char *filename, 
			STAMNES_TABLE **table)
{
  int i=0, j=0, status=0;
  int rows=0, columns=0, max_columns=0;
  double **value=NULL;


  /* read file to two-dimensional double array value */
  status = ASCII_file2double (filename,
			      &rows, &max_columns, &columns,
			      &value);

  if (status != 0)
    return status;


  /* check if a rectangular matrix */  
  if (columns != max_columns)  { 
    ASCII_free_double (value, rows);
    fprintf (stderr, " ... read_stamnes_table(): %s is not a rectangular matrix!\n", filename);
    fprintf (stderr, " ... minimum number of columns: %d, maximum number of columns: %d\n", 
	     columns, max_columns);

    return NOT_A_VALID_STAMNES_TABLE;
  }
    

  /* allocate memory for Stamnes table */
  *table = calloc (1, sizeof(STAMNES_TABLE));

  
  /* first row/column contains ratio/zenith */
  (*table)->n_zenith = columns-1;
  (*table)->n_ratio  = rows-1;
 
  /* allocate memory for arrays */
  (*table)->zenith = (double *)  calloc ((*table)->n_zenith, sizeof (double));
  (*table)->ratio  = (double *)  calloc ((*table)->n_ratio,  sizeof (double));
  (*table)->ozone  = (double **) calloc ((*table)->n_ratio,  sizeof (double *));

  for (i=0; i<(*table)->n_ratio; i++)
    (*table)->ozone[i] = (double *) calloc ((*table)->n_zenith,  sizeof (double));
  
  /* copy first row to array zenith */
  for (i=1; i<=(*table)->n_zenith; i++)  
    (*table)->zenith[i-1] = value[0][i];

  /* copy first column to array ratio */
  for (j=1; j<=(*table)->n_ratio; j++)  
    (*table)->ratio[j-1] = value[j][0];



  /* copy the remainder to 2dim double array ozone */
  for (i=1; i<=(*table)->n_ratio; i++) 
    for (j=1; j<=(*table)->n_zenith; j++)
      (*table)->ozone[i-1][j-1] = value[i][j];
    

  /* free memory of double array */
  ASCII_free_double (value, rows);
 
  return 0;  /* if o.k. */
}





/**************************************************************/
/* free memory of data read from Stamnes table.               */
/**************************************************************/

void free_stamnes_table (STAMNES_TABLE *table)
{
  int i=0;

  for (i=0; i<table->n_ratio; i++) 
    free (table->ozone[i]);

  free(table->ozone);

  free(table->zenith);
  free(table->ratio);

  free(table);
}



/***************************************************************/
/* stamnes_calculate uses the data read from a Stamnes table   */
/* to calculate the ozone column for given zenith angle and    */
/* irradiance ratio.                                           */
/*                                                             */
/* zenith[0..n_zenith-1]  zenith angles found in Stamnes table */
/* ratio[0..n_ratio-1]    ratio's found in Stamnes table       */
/* ozone[0..n_ratio-1][0..n_zenith-1]   ozone columns found in */
/*   Stamnes table.                                            */
/* meas_zenith   zenith angle to be used for calculation       */
/* meas_ratio    ratio to be used for calculation              */
/* column                                                      */
/***************************************************************/

int stamnes_calculate (STAMNES_TABLE *table,
		       double meas_zenith, double meas_ratio,
		       double *column)
{
  int counter=0, status=0;
  double *a0=NULL, *a1=NULL, *a2=NULL, *a3=NULL;
  double *b0=NULL, *b1=NULL, *b2=NULL, *b3=NULL;
  double lower=0, upper=0;

  /* reset ozone column */
  *column = -1;

  /* look for neighbour points in  array of ratios */
  while (counter < table->n_ratio && table->ratio[counter] < meas_ratio)
    counter++;

  /* check if meas_ratio within Stamnes table limits */ 
  if (counter<=0 || counter >= table->n_ratio)
    return RATIO_OUT_OF_RANGE;


  /* calculate spline coefficients for ozone as function of zenith */
  status = spline_coeffc (table->zenith, table->ozone[counter-1], table->n_zenith, 
			  &a0, &a1, &a2, &a3);

  if (status!=0)  {
    free(a0);
    free(a1);
    free(a2);
    free(a3);
    return ERROR_SPLINING_ZENITH;
  }
    

  status = spline_coeffc (table->zenith, table->ozone[counter], table->n_zenith, 
			  &b0, &b1, &b2, &b3);
  
  if (status!=0)  {
    free(a0);
    free(a1);
    free(a2);
    free(a3);
    free(b0);
    free(b1);
    free(b2);
    free(b3);
    return ERROR_SPLINING_ZENITH;
  }

  /* calculate interpolated data */
  status =  calc_splined_value (meas_zenith, &lower, 
				table->zenith, table->n_zenith, 
				a0, a1, a2, a3);

  free(a0);
  free(a1);
  free(a2);
  free(a3);
  

  if (status!=0)  {
    free(b0);
    free(b1);
    free(b2);
    free(b3);
    return ERROR_SPLINING_ZENITH;
  }

  status =  calc_splined_value (meas_zenith, &upper, 
				table->zenith, table->n_zenith, 
				b0, b1, b2, b3);

  free(b0);
  free(b1);
  free(b2);
  free(b3);


  if (status!=0) 
    return ERROR_SPLINING_ZENITH;


  /* now interpolate linearly for ratio */
  *column = lower + (upper - lower) / (table->ratio[counter] - table->ratio[counter-1]) *
    (meas_ratio - table->ratio[counter-1]);

  return 0;  /* if o.k. */
}





